from munch import Munch
from datetime import datetime
import importlib.resources
from pathlib import Path
import random
import string
import os
import platform

Shared = Munch()

Shared.data = Munch()

Shared.runtime = Munch()
Shared.runtime.base = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
with importlib.resources.path('cap', 'VERSION') as path:
    Shared.runtime.capVersion = Path(path).read_text()
Shared.runtime.capLog = 'ToBeSet'
Shared.runtime.hailLog = 'ToBeSet'
Shared.runtime.hailVersion = 'ToBeSet'
Shared.runtime.sparkVersion = 'ToBeSet'
Shared.runtime.sparkConfig = 'ToBeSet'
Shared.runtime.dateTime = str(datetime.now().strftime("%Y/%m/%d-%H:%M:%S"))
Shared.runtime.host = platform.node()
Shared.runtime.environment = Munch()

for envVar in ['CAP_DIR', 'PYSPARK_PYTHON', 'PYSPARK_PYTHON', 'HAIL_HOME', 'SPARK_HOME', 'HDFS_HOME']:
    if envVar in os.environ:
        Shared.runtime.environment[envVar] = os.environ[envVar]
    else:
        Shared.runtime.environment[envVar] = '__NOT_SET__'

Shared.fileSystem = None

Shared.vepCheckWaitTime = 5

Shared.CurrentStageForLogging = None
Shared.CurrentFunctionForLogging = list()

Shared.numPartitions = Munch()
Shared.numPartitions.default = 4
Shared.numPartitions.min = 1
Shared.numPartitions.max = 32  

Shared.numSgeJobs = Munch()
Shared.numSgeJobs.default = 4
Shared.numSgeJobs.min = 1
Shared.numSgeJobs.max = 32


from munch import Munch, munchify
import jsonschema

import hail as hl
from .logutil import *
from .common import *
from .helper import *
from .decorators import *
from .pyobj import PyObj
from .shared import Shared

if __name__ == '__main__':
    print('This module is not executable. Please import this module in your program.')
    exit(0)

#TBF add to be deleted flag to inout and delete files after sucessfull compeletion of the workload
class Workload(PyObj):

    @D_General
    def __init__(self, path, format=None, reset=False):
        super().__init__(path, format, reset)

        if reset:
            self.Clear()

        if 'globConfig' not in self.obj:
            self.obj.globConfig = Munch()
        if 'config' not in self.obj:
            self.obj.config = Munch()

        self.globConfig = self.obj.globConfig
        self.config = self.obj.config
        self.order = self.obj.order
        self.stages = self.obj.stages

        if 'runtimes' not in self.globConfig:
            self.globConfig.runtimes = list()
        self.globConfig.runtimes.append(Shared.runtime)
        self.Update()

        CheckShared()
        Shared.update(self.globConfig)
        CheckShared()
        
        if self.order:
            for stageId in self.order:
                if stageId not in self.stages:
                    LogException(f'{stageId} is listed in the "order" but not defined in "stages".')

        for stageId, stage in self.stages.items():
            if 'spec' not in stage:
                print(stage)
                LogException(f'stage {stageId} does not have the "spec"')
            
            # Push stage id and inout name into the structure
            stage.spec.id = stageId
            if 'inout' in stage:
                for name, inout in stage.inout.items():
                    inout.name = name

            if 'status' not in stage.spec:
                stage.spec.status = 'Initiated'

        self.stageSchema = PyObj(path='StageSchema.json', isInternal=True, isSchema=True)
        self.functionsSchema = PyObj(path='FunctionsSchema.json', isInternal=True)
        
        self.workloadSchema = PyObj(path='WorkloadSchema.json', isInternal=True, isSchema=True)
        self.schema = self.workloadSchema.obj
        self.CheckObject()

        self.Update()

        self.CheckStages()

        Log('Initialised')

    @D_General
    def Clear(self):
        obj = {
            'globConfig': dict(),
            'config': dict(),
            'order': list(),
            'stages': dict()
        }
        self.obj = munchify(obj)
        self.Update()
        Log('Cleared')

    @D_General
    def InferFileFormat(self, inout, name):

        def TestFormat(inout, name, suffix, format, compression):
            if inout.path.endswith(suffix):
                inout.format = format
                inout.compression = compression
                Log(f'<< inout: {name} >> Inferred Format:Compression is {format}:{compression}.')

        if inout.pathType == 'file' and 'format' not in inout:

            if 'compression' in inout:
                LogException(f'<< inout: {name} >> When format is not provided (infer format) compression should not be provided.')

            TestFormat(inout, name, '.mt', 'mt', 'None')
            TestFormat(inout, name, '.ht', 'ht', 'None')
            TestFormat(inout, name, '.vcf', 'vcf', 'None')
            TestFormat(inout, name, '.vcf.gz', 'vcf', 'gz')
            TestFormat(inout, name, '.vcf.bgz', 'vcf', 'bgz')
            TestFormat(inout, name, '.tsv', 'tsv', 'None')
            TestFormat(inout, name, '.tsv.gz', 'tsv', 'gz')
            TestFormat(inout, name, '.tsv.bgz', 'tsv', 'bgz')
            TestFormat(inout, name, '.csv', 'csv', 'None')
            TestFormat(inout, name, '.csv.gz', 'csv', 'gz')
            TestFormat(inout, name, '.csv.bgz', 'csv', 'bgz')
            TestFormat(inout, name, '.json', 'json', 'None')
            TestFormat(inout, name, '.json.gz', 'json', 'gz')
            TestFormat(inout, name, '.json.bgz', 'json', 'bgz')
            TestFormat(inout, name, '.bed', 'bed', 'None')
            TestFormat(inout, name, '.bim', 'bim', 'None')
            TestFormat(inout, name, '.fam', 'fam', 'None')

        if 'format' not in inout:
            LogException(f'<< inout: {name} >> Format is not provided and cannot be inferred.')

        if 'compression' not in inout:
            inout.compression = 'None'

    @D_General
    def CheckInout(self, stage):
        """Check all inout of the `stage`.

        Args:
            stage (Stage): the stage to be processed.
        """

        Log(f'There are {len(stage.inout)} inout/s to be checked.')

        for name, inout in stage.inout.items():
            Log(f'<< inout: {name} >> Checking...')
            inout.path = AbsPath(inout.path)
            
            if 'pathType' not in inout:
                inout.pathType = 'file'

            self.InferFileFormat(inout, name)

            if inout.format not in ['mt', 'ht']:
                if 'isAlive' in inout:
                    LogException(f'<< inout: {name} >> isAlive should not be presented when input format is {inout.format}')
            else:
                if 'isAlive' not in inout:
                    inout.isAlive = True
                if inout.isAlive:
                    ### TBF what if the user dont want to repartition at all
                    if 'numPartitions' not in inout:
                        inout.numPartitions = Shared.numPartitions.default

                    if not (Shared.numPartitions.min <= inout.numPartitions <= Shared.numPartitions.max):
                        LogException(f'<< inout: {name} >> numPartitions {inout.numPartitions} must be in range [{Shared.numPartitions.min}, {Shared.numPartitions.max}]')

                    for key in ['toBeCached', 'toBeCounted']:
                        if key not in inout:
                            inout[key] = True

            if 'isAlive' in inout and not inout.isAlive:
                for key in ['numPartitions', 'toBeCached', 'toBeCounted']:
                    if key in inout:
                        LogException(f'<< inout: {name} >> When isAlive is explicitly set to false, "{key}" should not be presented at inout.')

        # TBF this file existance check needs to be reviewd
        if stage.spec.status != 'Completed':
            for name, inout in stage.inout.items():
                if inout.direction == 'output':
                    if inout.format == 'bfile':
                        cond = any([FileExist(f'{inout.path}{suffix}') for suffix in ['bed', 'bim', 'fam']])
                    else:
                        cond = FileExist(inout.path)

                    if cond:
                        LogException(f'<< inout: {name} >> Output path (or plink bfile prefix) {inout.path} already exist in the file system')

    @D_General
    def CheckStage(self, stage):
        """Check if stage is ok to be executed.

        Note:
            - This function should be called just before executing the stage beacuase it checks if the output file exist and prevent overwriting before executing stage.

        Args:
            stage (Stage): the stage to be processed.
        """

        # Compelete stage schema template by adding arg and inout field for the specific function.
        try:
            functionsSchema = self.functionsSchema.obj
            stageSchema = self.stageSchema.obj

            for key in ['arg', 'inout']:
                if key not in functionsSchema[stage.spec.function]:
                    LogException(f'The FunctionSchema does not include {key} for {stage.spec.function}')
                stageSchema['properties'][key] = functionsSchema[stage.spec.function][key]

            jsonschema.Draft7Validator.check_schema(stageSchema)
            jsonschema.validate(instance=stage, schema=stageSchema)
        except jsonschema.exceptions.SchemaError:
            LogException(f'StageSchema is invalid')
        except jsonschema.exceptions.ValidationError:
            LogException(f'Stage is not validated by the schema')

        # Step 3: Check each Input/Output (inout).
        self.CheckInout(stage)

        LogPrint(f'Stage is Checked')

    @D_General
    def CheckStages(self):
        if self.order:
            for stageId in self.order:
                stage = self.stages[stageId]
                Shared.CurrentStageForLogging = stage
                self.CheckStage(stage)
                Shared.CurrentStageForLogging = None

    @D_General
    def ProcessLiveInput(self, input):
        Log(f'<< inout: {input.name} >> is {JsonDumps(input)}.')
        if 'isAlive' in input and input.isAlive:
            if input.path not in Shared.data:
                Log(f'<< inout: {input.name} >> Loading.')
                try:
                    if input.format == 'ht':
                        mht = hl.read_table(input.path)
                    elif input.format == 'mt':
                        mht = hl.read_matrix_table(input.path)
                    else:
                        pass  # Already handled in CheckInout
                except:
                    LogException(f'<< inout: {input.name} >> Cannot read input form {input.path}.')
                else:
                    Shared.data[input.path] = mht
                    Log('<< inout: {name} >> Loaded.')
            else:
                mht = Shared.data[input.path]
                Log(f'<< inout: {input.name} >> Preloaded.')

            if input.numPartitions and mht.n_partitions() != input.numPartitions:
                np = mht.n_partitions()
                mht = mht.repartition(input.numPartitions)
                Log(f'<< inout: {input.name} >> Repartitioned from {np} to {input.numPartitions}.')
            if input.toBeCached:
                mht = mht.cache()
                Log(f'<< inout: {input.name} >> Cached.')
            if input.toBeCounted:
                input.count = Count(mht)
                Log(f'<< inout: {input.name} >> Counted.')
                self.Update()

    @D_General
    def ProcessLiveInputs(self, stage):
        """Make sure that live input files are loaded in the shared object.

        Args:
            stage (Stage): To be Processed.
        """

        numInput = len([1 for inout in stage.inout.values() if inout.direction == 'input'])
        Log(f'Out of {len(stage.inout)} inouts {numInput} are inputs.')

        for input in stage.inout.values():
            if input.direction == 'input':
                self.ProcessLiveInput(input)

    @D_General
    def ProcessLiveOutput(self, output):
        
        Log(f'<< inout: {output.name} >> is {JsonDumps(output)}')

        if 'isAlive' in output and output.isAlive:
            if 'data' in output:
                mht = output.data
            else:
                LogException(f'<< inout: {output.name} >> No "data" field is provided.')

            if output.path in Shared.data:
                LogException(f'<< inout: {output.name} >> Output path {output.path} alredy exist in the shared.')

            if output.numPartitions and mht.n_partitions() != output.numPartitions:
                np = mht.n_partitions()
                mht = mht.repartition(output.numPartitions)
                Log(f'<< inout: {output.name} >> Repartitioned from {np} to {output.numPartitions}.')

            if output.toBeCached:
                mht = mht.cache()
                Log(f'<< inout: {output.name} >> Cached.')

            if output.toBeCounted:
                output.count = Count(mht)
                Log(f'<< inout: {output.name} >> Counted.')
                self.Update()

            Shared.data[output.path] = mht
            Log(f'<< inout: {output.name} >> Added to shared.')

            if output.format in ['ht', 'mt']:
                mht.write(output.path, overwrite=False)
                Log(f'<< inout: {output.name} >> Dumped.')
            else:
                pass  # Already handled in CheckInout

    @D_General
    def ProcessLiveOutputs(self, stage):
        """Write live output to disk.

        Args:
            stage (Stage): To be processed
        """

        numOutput = len([1 for inout in stage.inout.values() if inout.direction == 'output'])
        logger.info(f'Out of {len(stage.inout)} inouts {numOutput} are outputs')

        for output in stage.inout.values():
            if output.direction == 'output':
                self.ProcessLiveOutput(output)



import json
import yaml
from munch import munchify
import jsonschema
import importlib.resources

from .logutil import *
from .common import *
from .helper import *
from .decorators import *

if __name__ == '__main__':
    print('This module is not executable. Please import this module in your program.')
    exit(0)


class PyObj:

    @D_General
    def __init__(self, path, format=None, reset=False, isInternal=False, isSchema=False, schema=None):

        self.path = path
        self.format = format
        self.obj = None
        self.isInternal = isInternal
        self.isSchema = isSchema
        self.schema = schema

        if isSchema and schema:
            Log(f'Schema object cannot have a schema')
        if (isInternal or isSchema):
            self.readOnly = True
            if reset:
                Log(f'Schema and Internal file are not supposed to be reset')
        else:
            self.readOnly = False

        if path.endswith('.json'):
            detectedFormat = 'json'
        elif path.endswith('.yaml'):
            detectedFormat = 'yaml'
        else:
            detectedFormat = None

        if not format:
            if detectedFormat:
                self.format = detectedFormat
                Log(f'Format is set to {detectedFormat}')
            else:
                LogException(f'Could not infer the file format.')
        else:
            if detectedFormat != format:
                LogException('Given format is different from file extension.')

        if not self.readOnly and (reset or not FileExist(self.path)):
            self.Clear()
        else:
            self.Load()

        if isSchema:
            self.CheckSchema()

        if schema:
            self.CheckObject()

        Log('Initialised.')

    @D_General
    def CheckSchema(self):
        try:
            jsonschema.Draft7Validator.check_schema(self.obj)
        except:
            LogException('Invalid schema')
        Log('Schema Validated.')

    @D_General
    def CheckObject(self):
        try:
            jsonschema.validate(instance=self.obj, schema=self.schema)
        except:
            LogException('Invalid Object based on schema')
        Log('Object Validated.')

    @D_General
    def LoadExtenal(self, path):
        if not FileExist(path):
            LogException(f'File does not exist: {path}')

        with open(path) as inFile:
            if self.format == 'json':
                obj = json.load(inFile)
            elif self.format == 'yaml':
                obj = yaml.load(inFile, Loader=yaml.FullLoader)
            self.obj = munchify(obj)

    @D_General
    def LoadInternal(self, name):
        with importlib.resources.path('cap', name) as path:
            self.LoadExtenal(str(path))

    @D_General
    def Load(self):
        if self.isInternal:
            self.LoadInternal(self.path)
        else:
            self.LoadExtenal(self.path)

        if self.schema:
            self.CheckObject()

        Log('Loaded')

    @D_General
    def UpdateExternal(self, path):
        obj = ToDict(self.obj)
        with open(path, 'w') as outfile:
            if self.format == 'json':
                json.dump(obj, outfile, indent=4, sort_keys=True, default=str)
            elif self.format == 'yaml':
                yaml.dump(obj, outfile, indent=4, sort_keys=True)

    @D_General
    def UpdateInternal(self, name):
        with importlib.resources.path('cap', name) as path:
            self.UpdateExternal(path)

    @D_General
    def Update(self):
        if self.readOnly:
            LogException('Is read-only')

        if self.schema:
            self.CheckObject()

        if self.isInternal:
            self.UpdateInternal(self.path)
        else:
            self.UpdateExternal(self.path)
        Log('Updated')

    @D_General
    def Clear(self):
        self.obj = dict()
        self.Update()
        Log('Cleared')



from .workload import Workload
from . import operation as Operation
from .helper import *
from .common import *
from .logutil import *
from .shared import Shared


import hail as hl
from munch import Munch
from datetime import datetime

from copy import deepcopy

if __name__ == '__main__':
    print('This module is not executable. Please import this module in your program.')
    exit(0)

class Executor:

    @D_General
    def __init__(self, workload, hailLog=None):

        if not isinstance(workload, Workload):
            LogException('workload must be of type Workload')
        self.workload = workload
        try:
            if hailLog:
                Shared.runtime.hailLog = hailLog
            else:
                randomStr = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
                now = str(datetime.now().strftime("%Y%m%d-%H%M%S"))
                Shared.runtime.hailLog = f'hail.{now}.{randomStr}.log'

            appName = f'CAP - {Shared.runtime.hailLog}'
     
            hl.init(log=Shared.runtime.hailLog, app_name=appName)

            Shared.runtime.hailVersion = hl.version()
            sc = hl.spark_context()
            Shared.runtime.sparkVersion = sc.version
            Shared.runtime.sparkConfig = dict(sc.getConf().getAll())
            workload.Update()
     

            LogPrint("+++++++++++++++++++++++++++++++")
            LogPrint("+++++++++++++++++++++++++++++++")
            LogPrint("+++++++++++++++++++++++++++++++")
            Log(f'Runtime Information: {Shared.runtime}')
            LogPrint("+++++++++++++++++++++++++++++++")
            LogPrint("+++++++++++++++++++++++++++++++")
            LogPrint("+++++++++++++++++++++++++++++++")
        except:
            LogException('Something wrong')

        self.initialised = True

    @D_General
    def Execute(self, reset=False):
        workload = self.workload
        if workload.order:
            for stageId in workload.order:
                stage = workload.stages[stageId]
                if stage.spec.status != 'Completed' or reset:
                    self.ExecuteStage(stage)

    @D_General
    def ExecuteStage(self, stage):
        workload = self.workload
        Shared.CurrentStageForLogging = stage
        workload.CheckStage(stage)  # Check the stage right before execution to make sure no dynamic error occurs
        LogPrint(f'Started')
        func = getattr(Operation, stage.spec.function)
        runtime = Munch()
        runtime.base = Shared.runtime.base
        if 'runtimes' not in stage.spec:
            stage.spec.runtimes = list()
        stage.spec.runtimes.append(runtime)
        runtime.startTime = datetime.now()
        workload.Update()
        workload.ProcessLiveInputs(stage)
        workload.Update()
        func(stage)
        workload.Update()
        workload.ProcessLiveOutputs(stage)
        workload.Update()
        runtime.endTime = datetime.now()
        runtime.execTime = str(runtime.endTime - runtime.startTime)
        runtime.status = 'Completed'
        stage.spec.status = 'Completed'
        workload.Update()
        LogPrint(f'Completed in {runtime.execTime}')
        Shared.CurrentStageForLogging = None




import time

import hail as hl
from hail.expr.expressions.expression_typecheck import T
from hail.methods.impex import export_bgen
from munch import Munch
from .logutil import *
from .common import *
from .helper import *
from .shared import Shared

if __name__ == '__main__':
    print('This module is not executable. Please import this module in your program.')
    exit(0)

# TBF: to include ref genome as parameter


@D_General
def ImportGenotype(stage):
    inout = stage.inout
    if 'arg' in stage:
        arg = stage.arg
    else:
        arg = Munch()

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outGt = inout.outGt

    # >>>>>>> Live Input <<<<<<<<

    # >>>>>>> STAGE Code <<<<<<<<
    # Loading input genotype
    try:
        if 'importParam' not in arg:
            arg.importParam = dict()
        if 'inputFormat' in arg:
            if arg.inputFormat != inGt.format:
                LogException(f'input format mentioned in arg {arg.inputFormat} is different from input format of the inout inGt {inGt.format}')

        if inGt.format == 'vcf':
            mt = hl.import_vcf(inGt.path, **arg.importParam)
        elif inGt.format == 'bfile':
            mt = hl.import_plink(bed=f'{inGt.path}.bed', bim=f'{inGt.path}.bim', fam=f'{inGt.path}.fam', **arg.importParam)
        else:
            LogException(f'inGt.format ({inGt.format}) is not supported')
    except:
        LogException('Hail cannot read genotype data.')
    Log(f'Genotypes are loaded from "{inGt.path}".')

    # Rename requested fields
    if 'rename' in arg:
        try:
            mt = mt.rename(arg.rename)
        except:
            LogException(f'Hail cannot rename MatrixTable using {arg.rename}.')
        Log(f'Renaming is done {arg.rename}.')

    # >>>>>>> Live Output <<<<<<<<
    outGt.data = mt


@D_General
def SplitMulti(stage):
    inout = stage.inout
    arg = stage.arg

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outGt = inout.outGt

    # >>>>>>> Live Input <<<<<<<<
    mt = Shared.data[inGt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    # Split multi allelic site
    try:
        before = Count(mt)
        if arg.withHTS:
            mt = hl.split_multi_hts(mt)
        else:
            mt = hl.split_multi(mt)
        after = Count(mt)
    except:
        LogException('Could not split multi allelic sites.')
    Log(f'{before.variants} loci result in {after.variants} bi-allelic loci.')
    # >>>>>>> Live Output <<<<<<<<
    outGt.data = mt


@D_General
def AddId(stage):
    inout = stage.inout
    arg = stage.arg

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outGt = inout.outGt
    outCol = inout.outCol
    outRow = inout.outRow

    # >>>>>>> Live Input <<<<<<<<
    mt = Shared.data[inGt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    # Add indexes
    mt = mt.annotate_rows(variantId=hl.str(':').join(hl.array([mt.locus.contig, hl.str(mt.locus.position)]).extend(mt.alleles)))
    mt = mt.annotate_cols(sampleId=mt[arg.sampleId])

    mt = mt.key_cols_by('sampleId')

    # extract row and col table and drop them from matrix table
    ht_col = mt.cols()
    ht_col = ht_col.key_by('sampleId')

    ht_row = mt.rows()
    ht_row = ht_row.key_by('variantId')

    dropKeys = (set(mt.row).union(set(mt.col)))-{'alleles', 'locus', 'sampleId', 'variantId'}
    mt = mt.drop(*list(dropKeys))

    # >>>>>>> Live Output <<<<<<<<
    outGt.data = mt
    outCol.data = ht_col
    outRow.data = ht_row


@D_General
def ExportGenotype(stage):
    inout = stage.inout
    if 'arg' in stage:
        arg = stage.arg
    else:
        arg = Munch()

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outGt = inout.outGt

    # >>>>>>> Live Input <<<<<<<<
    mt = Shared.data[inGt.path]

    # >>>>>>> STAGE Code <<<<<<<<

    # If epxorting for VEP Overwire all other parameters
    if 'forVep' in arg and arg.forVep:
        if 'drop' in arg:
            LogException('When exporting for VEP (forVep=True) drop argument must not presented')
        #TBF arg may not present
        # if arg.outputFormat != 'vcf' or outGt.format != 'vcf' or outGt.compression != 'bgz':
        #     LogException('When exporting for VEP (forVep=True) arg.outputFormat and outGt.format must be "vcf" and outGt.compression must be "bgz"')
        # if 'exportParam' not in arg or 'parallel' not in arg.exportParam or arg.exportParam.parallel != 'separate_header':
        #     LogException('When exporting for VEP (forVep=True) arg.exportParam.parallel must be set to "separate_header"')

        try:
            ht = mt.rows().select('variantId')
            mt = hl.MatrixTable.from_rows_table(ht)
            mt = mt.annotate_cols(sample='x')
            mt = mt.key_cols_by(mt.sample)
            mt = mt.annotate_rows(rsid=hl.str(mt.variantId))
        except:
            LogException('Could not extract variant-only MatrixTable for VEP')
    else:
        mt = mt.annotate_cols(ID=hl.str(mt.sampleId))
        mt = mt.key_cols_by(mt.ID)

    # Drop fileds as requested before export
    if 'drop' in arg:
        validFields = list(mt.row) + list(mt.row) + list(mt.globals) + list(mt.entry)
        for field in arg.drop:
            if field in validFields:
                try:
                    mt = mt.drop(field)
                except:
                    LogException(f'Cannot drop {field} form tables. Valid fields include {validFields}.')
        Log(f'Fields are droped from the MatrixTable {arg.drop}.')

    # Export Genotypes
    try:
        if 'exportParam' not in arg:
            arg.exportParam = dict()
        #if arg.outputFormat == 'vcf' and outGt.format == 'vcf':
        if outGt.format == 'vcf':
            hl.export_vcf(mt, outGt.path, **arg.exportParam)
        #elif arg.outputFormat == 'bfile' and outGt.format == 'bfile':
        elif outGt.format == 'bfile':
            for k in ['call', 'fam_id', 'ind_id', 'pat_id', 'mat_id', 'is_female', 'pheno', 'varid', 'cm_position']:
                if k in arg.exportParam:
                    arg.exportParam[k] = HailPath([mt]+arg.exportParam[k])
            hl.export_plink(mt, outGt.path, **arg.exportParam)
        else:
            LogException(f'Output format is not properly set. Make sure arg.outputFormat ({arg.outputFormat}) and outGt.format ({outGt.format}) are consistent.')

    except:
        LogException('Hail cannot write genotype data.')
    Log(f'Genotypes are stored from "{outGt.path}".')

    # >>>>>>> Live Output <<<<<<<<


@D_General
def ImportPhenotype(stage):
    spec, arg, inout = stage.spec, stage.arg, stage.inout

    # >>>>>>> Input/Output <<<<<<<<
    inPt = inout.inPt
    inS = inout.inS  # input samples with sample IDs
    outPt = inout.outPt

    # >>>>>>> Live Input <<<<<<<<
    htS = Shared.data[inS.path]

    # >>>>>>> STAGE Code <<<<<<<<
    try:
        if 'importParam' not in arg:
            arg.importParam = dict()

        if inPt.format in ['tsv', 'tsv.bgz', 'tsv.gz']:
            sep = '\t'
        elif inPt.format in ['csv', 'csv.bgz', 'csv.gz']:
            sep = ','
        else:
            pass  # Already handled in Above

        imputeFlag = True
        if 'types' in arg.importParam:
            imputeFlag = False

        ht = hl.import_table(paths=inPt.path, impute=imputeFlag, delimiter=sep, **arg.importParam)
        ht = ht.key_by(arg.phenoKey)
    except:
        LogException('Hail cannot read phenotype data')
    Log(f'Phenotyps are loaded from "{inPt.path}"')

    htS = htS.key_by(arg.sampleKey)
    htS = htS.select('sampleId')
    ht = ht.join(htS)

    Log('Sample ids are added.')

    # >>>>>>> Live Output <<<<<<<<
    outPt.data = ht


@D_General
def PcaHweNorm(stage):
    spec, arg, inout = stage.spec, stage.arg, stage.inout

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outPcaScore = inout.outPcaScore

    # >>>>>>> Live Input <<<<<<<<
    mt = Shared.data[inGt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    if 'minMaf' in arg:
        minMaf = arg.minMaf
        if not isinstance(minMaf, float) or not (0 <= minMaf <= 0.5):
            LogException('minMaf parameter is not valid.')
        Log(f'Filter for minor allele frequency with minimum threshold of {minMaf}')
        mt = mt.annotate_rows(maf=hl.min(hl.agg.call_stats(mt.GT, mt.alleles).AF))
        mt = mt.filter_rows(mt.maf >= minMaf, keep=True)

    if 'ldR2' in arg:
        ldR2 = arg.ldR2
        if not isinstance(ldR2, float) or not (0 <= ldR2 <= 0.5):
            LogException('ldR2 parameter is not valid.')
        Log(f'LD pruning with r2={ldR2}')
        prunList = hl.ld_prune(mt.GT, r2=ldR2)
        mt = mt.filter_rows(hl.is_defined(prunList[mt.row_key]))

    if 'subSample' in arg:
        subSample = arg.subSample
        mt = SampleRows(mt, subSample)

    try:
        cl = dict()
        if 'outPcaLoading' in inout:
            cl['compute_loadings'] = True
        else:
            cl['compute_loadings'] = False
        eigenvalues, pcs, loading = hl.hwe_normalized_pca(mt.GT, k=arg.numPcaVectors, **cl)
    except:
        LogException('Hail cannot perform the pca analysis')

    logger.info(f'PCA is computed')

    # >>>>>>> Live Output <<<<<<<<
    outPcaScore.data = pcs
    if 'outPcaEigen' in inout:
        inout.outPcaEigen.data = eigenvalues
    if 'outPcaLoading' in inout:
        inout.outPcaLoading.data = loading
    if 'outPcaVarList' in inout:
        inout.outPcaVarList.data = mt.rows().select()


@D_General
def CalcQC(stage):
    spec, arg, inout = stage.spec, stage.arg, stage.inout

    # >>>>>>> Input/Output <<<<<<<<
    inGt = inout.inGt
    outQc = inout.outQc

    # >>>>>>> Live Input <<<<<<<<
    mt = Shared.data[inGt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    try:
        if arg.axis == 'sample':
            mt = hl.sample_qc(mt, name='qc')
            ht = mt.cols().select('qc')
        elif arg.axis == 'variant':
            mt = hl.variant_qc(mt, name='qc')
            ht = mt.rows()
            ht = ht.key_by('variantId')
            ht = ht.select('qc')
        else:
            pass  # Already handled above
    except:
        LogException(f'Hail cannot compute QC metrics for {arg.axis}')
    Log(f'PCA is computed')

    # >>>>>>> Live Output <<<<<<<<
    outQc.data = ht


@D_General
def ToMySql(stage):
    spec, arg, inout = stage.spec, stage.arg, stage.inout

    # >>>>>>> Input/Output <<<<<<<<
    inHt = inout.inHt

    # >>>>>>> Live Input <<<<<<<<
    ht = Shared.data[inHt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    ht = FlattenTable(ht)
    try:
        #TBF overwrite? really?
        ht.to_spark().write.format('jdbc').options(**arg.mySqlConfig).mode('overwrite').save()
    except:
        LogException('Hail cannot write data into MySQL database')
    Log(f'Data is exported to MySQL')

    # >>>>>>> Live Output <<<<<<<<

@D_General
def ToText(stage):
    inout = stage.inout
    if 'arg' in stage:
        arg = stage.arg
    else:
        arg = Munch()

    # >>>>>>> Input/Output <<<<<<<<
    inHt = inout.inHt
    outText = inout.outText

    # >>>>>>> Live Input <<<<<<<<
    ht = Shared.data[inHt.path]

    # >>>>>>> STAGE Code <<<<<<<<
    ht = FlattenTable(ht)
    try:
        if 'exportParam' not in arg:
            arg.exportParam = dict()
        ht.export(outText.path, **arg.exportParam)
    except:
        LogException(f'Hail cannot write data into a file {outText.path}')
    Log(f'Data is exported to {outText.path}')

    # >>>>>>> Live Output <<<<<<<<


@D_General
def VepAnnotation(stage):
    spec, arg, inout = stage.spec, stage.arg, stage.inout

    # >>>>>>> Input/Output <<<<<<<<
    inVar = inout.inVar
    outData = inout.outData  # use the path for the output files

    # >>>>>>> Live Input <<<<<<<<

    # >>>>>>> STAGE Code <<<<<<<<
    try:
        path = inVar.path
        templateCommand = arg.vepCli

        Bash(command=['mkdir', outData.path], isPath=[False, True])

        vcfList = WildCardPath(path + '/part-*.bgz')
        numJob = len(vcfList)

        if 'isArrayJob' in arg and arg.isArrayJob:

            numSgeJobs = Shared.numSgeJobs

            if 'numSgeJobs' in arg:
                if not (numSgeJobs.min <= arg.numSgeJobs <= numSgeJobs.max):
                    LogException(f'numSgeJobs {arg.numSgeJobs} must be in range [{numSgeJobs.min},{numSgeJobs.max}]')
            else:
                arg.numSgeJobs = numSgeJobs.default

            # Get the absolute path to the scripts
            templateCommand[1] = AbsPath(templateCommand[1])
            templateCommand[10] = AbsPath(templateCommand[10])
            # submit the array job

            command = templateCommand
            command = [path if p == '__VCF_DIR__' else p for p in command]
            command = [outData.path if p == '__JSON_DIR__' else p for p in command]
            command = [outData.path if p == '__TBL_DIR__' else p for p in command]
            command = [outData.path if p == '__JOB_DIR__' else p for p in command]
            command = [f'CAP' if p == '__JOB_NAME__' else p for p in command]
            command = ['1' if p == '__JOB_START__' else p for p in command]
            command = [str(numJob) if p == '__JOB_END__' else p for p in command]
            command = [str(arg.numSgeJobs) if p == '__JOB_IN_PARALLEL__' else p for p in command]

            Bash(command, isPath=[False, True, True, True, True, True, False, False, False, False, True])
        else:
            # Get the absolute path to the scripts
            templateCommand[1] = AbsPath(templateCommand[1])
            templateCommand[7] = AbsPath(templateCommand[7])
            # submit a job for each VCF
            for vcf in vcfList:
                fileName = os.path.basename(vcf)
                code = fileName[5:10]
                command = templateCommand
                command = [vcf if p == '__IN_VCF__' else p for p in command]
                command = [os.path.join(outData.path, f'part-{code}.json.bgz')if p == '__OUT_JSON__' else p for p in command]
                command = [os.path.join(outData.path, f'part-{code}.table') if p == '__OUT_TBL__' else p for p in command]
                command = [os.path.join(outData.path, f'part-{code}.job') if p == '__OUT_JOB__' else p for p in command]
                command = [f'CAP-{code}' if p == '__JOB_ID__' else p for p in command]

                Bash(command, isPath=[False, True, True, True, True, False, True, True])

        LogPrint(f'All {numJob} jobs are submitted.')

        # Wait Until all jobs are compeleted
        passed = 0
        while True:
            numCompeleted = 0
            for vcf in vcfList:
                fileName = os.path.basename(vcf)
                code = fileName[5:10]
                doneFile = os.path.join(outData.path, f'part-{code}.job.done')
                if FileExist(doneFile, silent=True):
                    numCompeleted += 1
            if numCompeleted != numJob:
                time.sleep(Shared.vepCheckWaitTime)
                passed += Shared.vepCheckWaitTime
                LogPrint(f'{numCompeleted} out of {numJob} compeleted in {passed} second ...')
            else:
                break
        LogPrint(f'All {numJob} jobs are compeleted.')
    except:
        LogException(f'Can not extract vcf file list.')
    Log(f'VEP JSON and TBL files are created.')

    # >>>>>>> Live Output <<<<<<<<


@D_General
def VepLoadTables(stage):
    inout = stage.inout
    arg = stage.arg

    # >>>>>>> Input/Output <<<<<<<<
    inData = inout.inData

    # >>>>>>> Live Input <<<<<<<<

    # >>>>>>> STAGE Code <<<<<<<<
    if 'all' in arg.tables:
        arg.tables = ['var', 'clvar', 'freq', 'conseq']

    path = inData.path
    try:  # TBF it currently check if the folder exist or not. should find a way to check all parquet files
        if 'var' in arg.tables:
            tblList = AbsPath(path + '/part-*.var.parquet')
            htVar = ImportMultipleTable(tblList)

        if 'clvar' in arg.tables:
            tblList = AbsPath(path + '/part-*.clvar.parquet')
            htClVar = ImportMultipleTable(tblList, addFileNumber=True)

        if 'freq' in arg.tables:
            tblList = AbsPath(path + '/part-*.freq.parquet')
            htFreq = ImportMultipleTable(tblList)

        if 'conseq' in arg.tables:
            tblList = AbsPath(path + '/part-*.conseq.parquet')
            htConseq = ImportMultipleTable(tblList)
    except:
        LogException(f'Can not read parquet files')

    try:
        if 'clvar' in arg.tables:
            # Process colocated-variants table
            htClVar = htClVar.annotate(clVarId=(htClVar.fileNumber * 2**32) + htClVar.clVarId)
        
        if 'conseq' in arg.tables:
            # Process consequences table and group consequences
            # Select all columns except 'varId' to group by (remove duplicate)
            groupBykeys = list(dict(htConseq.row).keys())
            groupBykeys.remove('varId')
            # Put all 'varId' for each uniq consequence into an array
            htConseq = htConseq.group_by(*list(groupBykeys)).aggregate(varIds=hl.agg.collect(htConseq.varId))
            htConseq = htConseq.add_index('conseqId')
            htConseq = htConseq.key_by('conseqId')
            # create a table for many to many relationship between consequences and variants
            htConseqToVar = htConseq.select('varIds')
            htConseqToVar = htConseqToVar.explode('varIds')
            htConseqToVar = htConseqToVar.rename({'varIds': 'varId'})
            # create a table to list consequence terms per consequence
            htConseqTerms = htConseq.select('consequence_terms', 'varIds')
            htConseqTerms = htConseqTerms.annotate(consequence_terms=hl.array(htConseqTerms.consequence_terms.replace('\[', '').replace('\]', '').replace('\'', '').split(',')))
            htConseqTerms = htConseqTerms.explode('consequence_terms')
            # create a table to list consequence terms per variant (this is for performance only but can be done by linking terms to consequences and then consequences to the variant).
            htVarTerms = htConseqTerms.explode('varIds')
            htVarTerms = htVarTerms.rename({'varIds': 'varId'})
            # drop un-neccessary field of each table
            htConseqTerms = htConseqTerms.select('consequence_terms')
            htVarTerms = htVarTerms.key_by('varId')
            htVarTerms = htVarTerms.select('consequence_terms')
            htConseq = htConseq.drop('consequence_terms', 'varIds')
    except:
        LogException('Cannot process tables.')

    Log(f'VEP parquet files are converted to hail tables.')

    # >>>>>>> Live Output <<<<<<<<
    if 'var' in arg.tables:
        inout.outVar.data = htVar
    if 'clvar' in arg.tables:
        inout.outClVar.data = htClVar
    if 'freq' in arg.tables:
        inout.outFreq.data = htFreq
    if 'conseq' in arg.tables:
        inout.outConseq.data = htConseq
        inout.outConseqToVar.data = htConseqToVar
        inout.outConseqTerms.data = htConseqTerms
        inout.outVarTerms.data = htVarTerms



import argparse
from os import path
from posixpath import abspath

from yaml import compose_all

from .logutil import *
from .common import *
from .decorators import *
from cap.workload import Workload
from cap.executor import Executor


@D_General
def CapMain(args):
    workload = Workload(path=args.workload)
    executor = Executor(workload, hailLog=args.hailLog)
    executor.Execute()


def Main():
    parser = argparse.ArgumentParser(
        description='Execute CAP workload'
    )
    parser.add_argument('-w', '--workload', required=True, type=str, help='The workload file (yaml or json).')
    parser.add_argument('-l', '--capLog', type=str, help='CAP log file')
    parser.add_argument('-hl', '--hailLog', type=str, help='Hail log file')
    args = parser.parse_args()
 
    if args.capLog:
        args.capLog = os.path.abspath(args.capLog)

    if args.hailLog:
        args.hailLog = os.path.abspath(args.hailLog)

    InitLogger(capLog=args.capLog)
    Log(f'Runtime Information: {Shared.runtime}')
    
    CapMain(args)


if __name__ == '__main__':
    Log('Collect logs for Main module.')
    Main()



import setuptools
from pathlib import Path
import subprocess

def GetVersion():

    version = Path('cap/VERSION').read_text().split('-')[0]
    commit = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
    commit = str(commit, "utf-8").strip()
    versionCommit = f'{version}-{commit}'
    Path('cap/VERSION').write_text(versionCommit)
    return version

setuptools.setup(
    name='cap-genomics',
    version=GetVersion(),
    description='Cohort Analysis Platform',
    long_description=Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    url='https://github.com/ArashLab/CAP',
    author='Arash Bayat',
    author_email='a.bayat@garvan.org.au',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    include_package_data=True,
    install_requires=['hail', 'munch', 'jsonschema', 'pyarrow', 'fastparquet'],
    packages=setuptools.find_packages()
)


from .common import *
from .logutil import *
from .shared import Shared

from datetime import datetime

if __name__ == '__main__':
    print('This module is not executable. Please import this module in your program.')
    exit(0)


def D_General(func):
    def W_General(*args, **kwargs):
        Log(f'Starts with args {JsonDumps(locals())}.', level='DEBUG')
        start = datetime.now()
        try:
            ret = func(*args, **kwargs)
        except:
            LogException('Unknown exception occurs.')
        end = datetime.now()
        Log(f'Ends in {end - start}', level='DEBUG')
        return ret
    return W_General
